/*
 * File......: DWETRAP.prg
 * Author....: Patrick O'Beirne, Systems Modelling Ltd.
 * CIS ID....: 100023,1304
 * Date......: $Date:   02 Nov 1991 21:26:00  $
 * Revision..: $Revision:   1.0  $
 * Log file..: $Logfile:                             $
 * 
 * This is an original work by Patrick O'Beirne and is placed in the
 * public domain.
 *
 * Modification history:
 * ---------------------
 *    Rev  1.2 Tue  01-10-1995  Just a better demo
 *    Rev  1.1 Tue  11-26-1991  11:36:06
 * Return numeric value rather than displaying Dialog()
 * Calls FT_XTOY rather than own etox() function
 * Allows numeric field ordinal position as first parameter
 * as well as character field name.
 *
 */


/*  $DOC$
 *  $FUNCNAME$
 *     FT_DWETRAP()
 *  $CATEGORY$
 *     Database
 *  $ONELINER$
 *     Field replace with data width/type error prevention
 *  $SYNTAX$
 *     FT_DWETRAP( <c/nFld_Name>, <xFld_Value> ) -> nErrCode
 *  $ARGUMENTS$
 *     <cFld_Name> is the Char name of the field to replace in
 * OR  <nFld_posn> is the ordinal position of the field in the record
 *     <xFld_Value> is the value (any type) to put in
 *  $RETURNS$
 *      0 if success, an error number if prevented by check.
 *  $DESCRIPTION$
 *     This function will check the width and type of xFld_Value
 *     against the database structure definition of cFld_Name.
 *     If the value fits, it is replaced, else an error code is returned.
 *     Note: calls FT_XTOY() which is also in NANFORUM library
 *  $SEEALSO$
 *    FT_XTOY() if you wish to force type conversion before
 *    doing the replace.
 *
 *  $EXAMPLES$
 *    In your code : 
 *     nErrCode := Personnel->(FT_DWETRAP('SALARY',123456.89))
 *
 *     Compile with /m /n /w /dFT_TEST to run the demo below
 *  $END$
 */

#include "dbstruct.ch"
#include "error.ch"

#ifdef FT_TEST
  Procedure MAIN( cTrap ) 
  // Pass ANY parameter to enable the trap
  // Pass none to let the Clipper error handler take it.
   local oErr
     
   // Create a dummy file to test the replaces
   dbcreate('DWETRAP.TMP', { {'Test1','C',6,0} ,{'Test2','N',3,0} })
   use dwetrap.tmp alias test
   append blank
   // Initialize with sample data
   field->Test1 := 'TEST1'
   field->Test2 := 123
   
   // Try replacing Numeric into char field
   if cTrap == NIL
      ? "About to test Data type error"
      Begin Sequence
         Replace field->Test1 with 1234
      Recover using oErr
         ? "Recovered "
         if oErr:classname=="ERROR" // This works in Clipper 5.2, not 5.01
            ?? "Error was",oErr:gencode,oErr:description
         else
            ?? "oErr is",FT_XTOY(oErr, "C")
         endif
      End Sequence
      ? "After test of Data type error"
   else
      test->(ft_testrap('Test1',1234))
   endif
   
   // Try replacing Numeric 4 digits into 3 digit field
   if cTrap == NIL
      ? "About to test Data width error"
      Begin Sequence
         Replace field->Test2 with 1234
      Recover using oErr
         ? "Recovered "
         if oErr:classname=="ERROR" // This works in Clipper 5.2, not 5.01
            ?? "Error was",oErr:gencode,oErr:description
         else
            ?? "oErr is",FT_XTOY(oErr, "C")
         endif
      End Sequence
      ? "After test of Data width error"
   else
      test->(ft_testrap('Test2',1234))
   endif
   
  return

   // FT_TesTrap puts a dialog around the replace trap for a demo
  static function FT_TESTRAP(cFld_name,xFld_Value)
   local nRet_val
   nRet_val := FT_DWETRAP(cFld_name,xFld_Value)
   do case
      case nRet_val == EG_NOVAR
         Dialog('Field '+FT_xtoy(cFld_Name,'C')+' is not in the current database "'+alias()+'"')
      case nRet_val == EG_DATATYPE
         Dialog('Field '+FT_xtoy(cFld_Name,'C')+' is type "'+type(cFld_name)+'"';
         +';and cannot take a value of '+FT_xtoy(xFld_Value,'C'))
      case nret_val == EG_DATAWIDTH
          Dialog('Field '+FT_xtoy(cFld_Name,'C')+' cannot take a value of '+ltrim(str(xFld_Value)))
   endcase
  return (nRet_val)

#endif   // def FT_TEST

function FT_DWETRAP(xField,xFld_Value)   // Replace with type and width check

static n10s :={1,10,100,1000,10000,100000,1000000,10000000,100000000,1000000000;
         ,10000000000,100000000000,1000000000000,10000000000000,100000000000000;
     ,1000000000000000,10000000000000000,100000000000000000,1000000000000000000;
     ,10000000000000000000,100000000000000000000}  // 10^0 .. 10^20
     // 10,000s lookups of a static array takes .5s vs 15s for 10^calc
local nFld_pos, cFld_name, aFld_stru := {}
local nFld_Max, nFld_Min, nFld_int, nRet_val := 0
// First parameter can be Numeric (field posn) or Char (field name)
if valtype(xField)=="N"
   nFld_pos := if(xField < 0 .or. xField > fcount(), 0, xField)
   cFld_name := field(nFld_pos)
elseif valtype(xField)=="C"
   cFld_name := xField
   nFld_pos := fieldpos(cFld_name)
else
   nFld_pos := 0
   nRet_val := EG_NOVAR
endif

if nFld_pos > 0
   aFld_stru := dbstruct()[nFld_pos]
   if aFld_stru[DBS_TYPE] == valtype(xFld_Value)
      if aFld_stru[DBS_TYPE] == 'N'
         nFld_int :=  aFld_stru[DBS_LEN] ;
                     -if(aFld_stru[DBS_DEC]==0,0,aFld_stru[DBS_DEC]+1)
         nFld_Min := - (n10s[nFld_int  ]-1/n10s[aFld_stru[DBS_DEC]+1])
         nFld_Max :=   (n10s[nFld_int+1]-1/n10s[aFld_stru[DBS_DEC]+1])
         if xFld_value < nFld_Min .or. xFld_value > nFld_Max
            //Dialog('Field '+cFld_name+' has a size of '+replicate('9',nFld_int);
            // +if(aFld_stru[DBS_DEC]=0,'','.'+replicate('9',aFld_stru[DBS_DEC]));
            // +';and cannot take a value of '+ltrim(str(xFld_Value)))
            nRet_val := EG_DATAWIDTH
         else
            // eval(Fieldblock(cFld_name,xFld_Value))  // Numeric
            Fieldput(nFld_pos,xFld_Value)             // Faster
         endif
      else
            Fieldput(nFld_pos,xFld_Value)
      endif
   else
      //Dialog('Field '+cFld_name+' is type "'+aFld_stru[DBS_TYPE]+'"';
      //   +';and cannot take a value of '+FT_xtoy(xFld_Value,'C'))
      nRet_val := EG_DATATYPE
   endif
else
   // Dialog('Field '+cFld_name+' is not in the current database "'+alias()+'"')
   nRet_val := EG_NOVAR
endif
return (nRet_val)


